# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

import logging
from datetime import datetime
from typing import Iterator, Optional, Sequence

from .http import Http, TimeoutT
from .utils import date_to_str

_LOGGER = logging.getLogger(__name__)


class Media(Http):
    def factory_create(self) -> str:
        ret = self.command("mediaFileFind.cgi?action=factory.create")
        return ret.content.decode()

    def factory_close(self, factory_id: str) -> str:
        ret = self.command(
            f"mediaFileFind.cgi?action=factory.close&object={factory_id}"
        )
        return ret.content.decode()

    def factory_destroy(self, factory_id: str) -> str:
        ret = self.command(
            f"mediaFileFind.cgi?action=factory.destroy&object={factory_id}"
        )
        return ret.content.decode()

    def media_file_find_start(
        self,
        factory_id: str,
        start_time: datetime,
        end_time: datetime,
        channel: int = 0,
        directories: Sequence[str] = (),
        types: Sequence[str] = (),
        flags: Sequence[str] = (),
        events: Sequence[str] = (),
        stream: Optional[str] = None,
    ) -> str:
        """
        https://s3.amazonaws.com/amcrest-files/Amcrest+HTTP+API+3.2017.pdf

        factory_id : returned by factory_create()

        dir : in which directories you want to find the file. It is an array.
                The index starts from 1. The range of dir is {"/mnt/dvr/sda0",
                "/mnt/dvr/sda1"}. This condition can be omitted. If omitted,
                find files in all the directories.

        type : which types of the file you want to find. It is an array. The
                index starts from 0. The range of type is {"dav", "jpg", "mp4"}
                If omitted, find files with all the types.

        flag : which flags of the file you want to find. It is an array. The
                index starts from 0. The range of flag is {"Timing", "Manual",
                "Marker", "Event", "Mosaic", "Cutout"}. If omitted, find files
                with all the flags.

        event : by which event the record file is triggered. It is an array.
                The index starts from 0. The range of event is {"AlarmLocal",
                "VideoMotion", "VideoLoss", "VideoBlind", "Traffic*"}. This
                condition can be omitted. If omitted, find files of all the
                events. stream : which video stream type you want to find.
                The range of stream is {"Main", "Extra1", "Extra2", "Extra3"}.
                If omitted, find files with all the stream types.
        """

        c_dirs = "".join(
            f"&condition.Dirs[{k}]={v}" for k, v in enumerate(directories)
        )

        c_types = "".join(
            f"&condition.Types[{k}]={v}" for k, v in enumerate(types)
        )

        c_flag = "".join(
            f"&condition.Flag[{k}]={v}" for k, v in enumerate(flags)
        )

        c_events = "".join(
            f"&condition.Events[{k}]={v}" for k, v in enumerate(events)
        )

        c_vs = f"&condition.VideoStream={stream}" if stream else ""

        start = date_to_str(start_time)
        end = date_to_str(end_time)
        extra_cond = "".join([c_dirs, c_types, c_flag, c_events, c_vs])
        ret = self.command(
            f"mediaFileFind.cgi?action=findFile&object={factory_id}&"
            f"condition.Channel={channel}&condition.StartTime={start}&"
            f"condition.EndTime={end}{extra_cond}"
        )
        return ret.content.decode()

    def media_file_find_next(self, factory_id: str, count: int = 100) -> str:
        ret = self.command(
            "mediaFileFind.cgi?action=findNextFile&"
            f"object={factory_id}&count={count}"
        )

        return ret.content.decode()

    def find_files(
        self,
        start_time: datetime,
        end_time: datetime,
        channel: int = 0,
        directories: Sequence[str] = (),
        types: Sequence[str] = (),
        flags: Sequence[str] = (),
        events: Sequence[str] = (),
        stream: Optional[str] = None,
    ) -> Iterator[str]:
        """
        https://s3.amazonaws.com/amcrest-files/Amcrest+HTTP+API+3.2017.pdf

        dir : in which directories you want to find the file. It is an array.
                The index starts from 1. The range of dir is {"/mnt/dvr/sda0",
                "/mnt/dvr/sda1"}. This condition can be omitted. If omitted,
                find files in all the directories.

        type : which types of the file you want to find. It is an array. The
                index starts from 0. The range of type is {"dav", "jpg", "mp4"}
                If omitted, find files with all the types.

        flag : which flags of the file you want to find. It is an array. The
                index starts from 0. The range of flag is {"Timing", "Manual",
                "Marker", "Event", "Mosaic", "Cutout"}. If omitted, find files
                with all the flags.

        event : by which event the record file is triggered. It is an array.
                The index starts from 0. The range of event is {"AlarmLocal",
                "VideoMotion", "VideoLoss", "VideoBlind", "Traffic*"}. This
                condition can be omitted. If omitted, find files of all the
                events. stream : which video stream type you want to find.
                The range of stream is {"Main", "Extra1", "Extra2", "Extra3"}.
                If omitted, find files with all the stream types.
        """
        factory_id = self.factory_create().strip().split("=")[1]
        _LOGGER.debug("%s findFile for factory_id=%s", self, factory_id)

        search = self.media_file_find_start(
            factory_id=factory_id,
            start_time=start_time,
            end_time=end_time,
            channel=channel,
            directories=directories,
            types=types,
            flags=flags,
            events=events,
            stream=stream,
        )

        if "ok" in search.lower():
            while True:
                _LOGGER.debug("%s findNextFile", self)
                content = self.media_file_find_next(factory_id)

                # The first line is 'found=N'.
                # However, it can be 'Error' if e.g. no more files found
                tag, _, str_count = content.split("\r\n", 1)[0].partition("=")
                if tag == "found":
                    count = int(str_count)
                else:
                    break

                _LOGGER.debug("%s returned %s %s", self, tag, count)

                if count == 0:
                    break

                yield content

            self.factory_close(factory_id)
            self.factory_destroy(factory_id)
        else:
            _LOGGER.debug("%s returned error: %s", self, search)

    def download_file(
        self,
        file_path: str,
        timeout: TimeoutT = None,
        stream: bool = False,
    ) -> bytes:
        """
        file_path: File location like returned by FilePath from find_files()
                   Example: /mnt/sd/2019-12-31/001/dav/00/00.12.00-00.20.00.mp4
        timeout:   Use default if None
        stream:    If True use streaming download instead of
                   reading content into memory
        """
        ret = self.command(
            "RPC_Loadfile/{0}".format(file_path),
            timeout_cmd=timeout,
            stream=stream,
        )
        return ret.content

    def download_time(
        self,
        start_time: datetime,
        end_time: datetime,
        channel: int = 0,
        stream: int = 0,
    ) -> bytes:
        """
        start_time and end_time are formatted as yyyy-mm-dd hh:mm:ss
        '%Y-%m-%d%%20%H:%M:%S'
        """
        start = date_to_str(start_time)
        end = date_to_str(end_time)
        ret = self.command(
            f"loadfile.cgi?action=startLoad&channel={channel}&"
            f"startTime={start}&endTime={end}&subtype={stream}"
        )
        return ret.content



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

import re
from typing import List, Optional, Tuple
from typing_extensions import TypedDict

from .http import Http
from .utils import percent, to_unit

_USED = ".UsedBytes"
_TOTAL = ".TotalBytes"


class StorageT(TypedDict):
    used_percent: str
    used: Tuple[str, str]
    total: Tuple[str, str]


class Storage(Http):
    @property
    def storage_device_info(self) -> str:
        ret = self.command("storageDevice.cgi?action=getDeviceAllInfo")
        return ret.content.decode()

    @property
    async def async_storage_device_info(self) -> str:
        ret = await self.async_command(
            "storageDevice.cgi?action=getDeviceAllInfo"
        )
        return ret.content.decode()

    @property
    def storage_device_names(self) -> str:
        ret = self.command("storageDevice.cgi?action=factory.getCollect")
        return ret.content.decode()

    @property
    async def async_storage_device_names(self) -> str:
        ret = await self.async_command(
            "storageDevice.cgi?action=factory.getCollect"
        )
        return ret.content.decode()

    @property
    def storage_used(self) -> Tuple[str, str]:
        (used,) = self._get_storage_values(self.storage_device_info, _USED)
        return to_unit(used)

    @property
    async def async_storage_used(self) -> Tuple[str, str]:
        (used,) = self._get_storage_values(
            await self.async_storage_device_info, _USED
        )
        return to_unit(used)

    @property
    def storage_total(self) -> Tuple[str, str]:
        (total,) = self._get_storage_values(self.storage_device_info, _TOTAL)
        return to_unit(total)

    @property
    async def async_storage_total(self) -> Tuple[str, str]:
        (total,) = self._get_storage_values(
            await self.async_storage_device_info, _TOTAL
        )
        return to_unit(total)

    @property
    def storage_used_percent(self) -> str:
        return self.storage_all["used_percent"]

    @property
    async def async_storage_used_percent(self) -> str:
        storage_all = await self.async_storage_all
        return storage_all["used_percent"]

    @property
    def storage_all(self) -> StorageT:
        used, total = self._get_storage_values(
            self.storage_device_info, _USED, _TOTAL
        )
        return self._build_storage_type(used, total)

    @property
    async def async_storage_all(self) -> StorageT:
        used, total = self._get_storage_values(
            await self.async_storage_device_info, _USED, _TOTAL
        )
        return self._build_storage_type(used, total)

    def _get_storage_values(self, info: str, *params) -> List[Optional[float]]:
        ret: List[Optional[float]] = []
        for param in params:
            match = re.search(f".{param}=([0-9.]+)", info)
            if match is None:
                ret.append(None)
            else:
                ret.append(float(match.group(1)))
        return ret

    def _build_storage_type(
        self, used: Optional[float], total: Optional[float]
    ) -> StorageT:
        if used is None or total is None:
            return {
                "used_percent": "unknown",
                "used": to_unit(used),
                "total": to_unit(total),
            }
        try:
            used_percent = percent(float(used), float(total))
        except (TypeError, ValueError, ZeroDivisionError):
            used_percent = "unknown"
        return {
            "used_percent": used_percent,
            "used": to_unit(used),
            "total": to_unit(total),
        }



"""
amcrest.exceptions

This module contains the set of amcrest's exceptions.
"""


class AmcrestError(Exception):
    """General Amcrest error occurred."""


class ReadTimeoutError(AmcrestError):
    """A read timeout error occurred."""


class CommError(AmcrestError):
    """A communication error occurred."""


class LoginError(AmcrestError):
    """A login error occurred."""



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import logging
import shutil
from typing import Optional

from urllib3.exceptions import HTTPError
from . import utils
from .exceptions import CommError
from .http import Http

_LOGGER = logging.getLogger(__name__)


class Audio(Http):
    @property
    def audio_input_channels_numbers(self) -> str:
        ret = self.command("devAudioInput.cgi?action=getCollect")
        return ret.content.decode()

    @property
    async def async_audio_input_channels_numbers(self) -> str:
        ret = await self.async_command("devAudioInput.cgi?action=getCollect")
        return ret.content.decode()

    @property
    def audio_output_channels_numbers(self) -> str:
        ret = self.command("devAudioOutput.cgi?action=getCollect")
        return ret.content.decode()

    @property
    async def async_audio_output_channels_numbers(self) -> str:
        ret = await self.async_command("devAudioOutput.cgi?action=getCollect")
        return ret.content.decode()

    def play_wav(
        self,
        httptype: Optional[str] = None,
        channel: Optional[int] = None,
        path_file: Optional[str] = None,
        encoding: str = "G.711A",
    ) -> None:

        if httptype is None:
            httptype = "singlepart"

        if channel is None:
            channel = 1

        if path_file is None:
            raise RuntimeError("filename is required")

        self.audio_send_stream(httptype, channel, path_file, encoding)

    def audio_send_stream(
        self,
        httptype: Optional[str] = None,
        channel: Optional[int] = None,
        path_file: Optional[str] = None,
        encode: Optional[str] = None,
    ) -> None:
        """
        Params:

            path_file - path to audio file
            channel: - integer
            httptype - type string (singlepart or multipart)

                singlepart: HTTP content is a continuos flow of audio packets
                multipart: HTTP content type is multipart/x-mixed-replace, and
                           each audio packet ends with a boundary string

            Supported audio encode type according with documentation:
                PCM
                ADPCM
                G.711A
                G.711.Mu
                G.726
                G.729
                MPEG2
                AMR
                AAC

        """
        if httptype is None or channel is None:
            raise RuntimeError("Requires htttype and channel")
        if encode is None:
            raise RuntimeError("Requires encode")
        if path_file is None:
            raise RuntimeError("Requires path_file")

        header = {
            "content-type": "Audio/" + encode,
            "content-length": "9999999",
        }

        cmd = (
            f"audio.cgi?action=postAudio&httptype={httptype}&channel={channel}"
        )
        with open(path_file, "rb") as f:
            file_audio = {"file": f}
            self.command_audio(
                cmd,
                file_content=file_audio,
                http_header=header,
            )

    def audio_stream_capture(
        self,
        httptype: Optional[str] = None,
        channel: Optional[int] = None,
        path_file: Optional[str] = None,
    ) -> bytes:
        """
        Params:

            httptype - type string (singlepart or multipart)
                singlepart: HTTP content is a continuos flow of audio packets
                multipart: HTTP content type is multipart/x-mixed-replace, and
                           each audio packet ends with a boundary string
            channel - integer
            path_file - path to output file
        """
        if httptype is None and channel is None:
            raise RuntimeError("Requires htttype and channel")

        ret = self.command(
            f"audio.cgi?action=getAudio&httptype={httptype}&channel={channel}",
            stream=True,
        )

        if path_file:
            try:
                with open(path_file, "wb") as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug(
                    "%s Audio stream capture to file failed due to error: %s",
                    self,
                    repr(error),
                )
                raise CommError(error) from error

        return ret.raw

    @property
    def audio_enabled(self) -> bool:
        """Return if any audio stream enabled."""
        return self.is_audio_enabled()

    @audio_enabled.setter
    def audio_enabled(self, enable: bool) -> None:
        """Enable/disable all audio streams."""
        self.set_audio_enabled(enable)

    @property
    async def async_audio_enabled(self) -> bool:
        """Return if any audio stream enabled."""
        return await self.async_is_audio_enabled()

    def is_audio_enabled(
        self, *, channel: int = 0, stream: str = "Main", stream_type: int = 0
    ) -> bool:
        """Return if the audio stream is enabled on the given channel.

        The stream should be either "Main", "Extra", or "Snap".  For the main
        stream, the stream type selects if it should read regular (0), motion
        detection (1), alarm (2), or emergency (3) encode settings.  The snap
        stream has the same settings for regular, motion detection, and alarm
        stream types.  For the extra stream, the stream type selects which
        stream should be read, and is zero indexed from 0 to 2 for streams 1 to
        3.
        """
        is_enabled = utils.extract_audio_video_enabled(
            f"{stream}Format[{stream_type}].Audio",
            self.encode_media,  # type: ignore[attr-defined]
        )
        return is_enabled[channel]

    async def async_is_audio_enabled(
        self, *, channel: int = 0, stream: str = "Main", stream_type: int = 0
    ) -> bool:
        """Return if any audio stream enabled on the given channel."""
        is_enabled = utils.extract_audio_video_enabled(
            f"{stream}Format[{stream_type}].Audio",
            await self.async_encode_media,  # type: ignore[attr-defined]
        )
        return is_enabled[channel]

    def set_audio_enabled(
        self, enable: bool, *, channel: int = 0, stream: str = "Main"
    ) -> None:
        """Enable/disable all audio streams on given channel."""
        self.command(
            utils.enable_audio_video_cmd(
                "Audio", enable, channel, stream=stream
            )
        )

    async def async_set_audio_enabled(
        self, enable: bool, *, channel: int = 0, stream: str = "Main"
    ) -> None:
        """Enable/disable all audio streams on given channel."""
        await self.async_command(
            utils.enable_audio_video_cmd(
                "Audio", enable, channel, stream=stream
            )
        )



"""Test log.py functions."""
import responses
import datetime

from .mocktestcase import MockTestCase


class TestLog(MockTestCase):
    """Tests for log.py."""

    @responses.activate
    def test_log_find(self):
        self.add_init_responses()

        responses.add(
            responses.GET,
            self.format_url('log.cgi', {
                'action': 'startFind',
                'condition.StartTime': '2020-01-02 21:05:33',
                'condition.EndTime': '2020-01-03 22:05:33'}),
            body='token=17\r\n',
            status=200)

        url_next = self.format_url('log.cgi', {
            'action': 'doFind',
            'token': 17,
            'count': 100})

        responses.add(
            responses.GET,
            url_next,
            body='\r\n'.join((
                'found=2',
                'items[0].Detail.Address=192.168.1.2',
                'items[0].Detail.Type=CGI',
                'items[0].Level=0',
                'items[0].Time=2020-01-02 22:05:33',
                'items[0].Type=Login',
                'items[0].User=admin',
                'items[1].Detail.Address=192.168.1.2',
                'items[1].Detail.Type=CGI',
                'items[1].Level=0',
                'items[1].Time=2020-01-02 22:06:33',
                'items[1].Type=Logout',
                'items[1].User=admin',
                '')),
            status=200)

        responses.add(
            responses.GET,
            url_next,
            body='\r\n'.join((
                'found=0',
                '')),
            status=200)

        responses.add(
            responses.GET,
            self.format_url('log.cgi', {
                'action': 'stopFind',
                'token': 17}),
            body='\r\n'.join((
                'OK',
                '')),
            status=200)

        c = self.get_amcrest().camera

        time_start = datetime.datetime(2020, 1, 2, 21, 5, 33)
        time_end = datetime.datetime(2020, 1, 3, 22, 5, 33)
        logs = list(c.log_find(time_start, time_end))
        self.assertEqual(2, len(logs))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from typing_extensions import Literal

from .http import Http

ActionT = Literal["start", "stop"]


class Ptz(Http):
    @property
    def ptz_config(self) -> str:
        return self._get_config("Ptz")

    @property
    async def async_ptz_config(self) -> str:
        return await self._async_get_config("Ptz")

    @property
    def ptz_auto_movement(self) -> str:
        return self._get_config("PtzAutoMovement")

    @property
    async def async_ptz_auto_movement(self) -> str:
        return await self._async_get_config("PtzAutoMovement")

    def ptz_presets_list(self, *, channel: int = 0) -> str:
        ret = self.command(f"ptz.cgi?action=getPresets&channel={channel}")
        return ret.content.decode()

    async def async_ptz_presets_list(self, *, channel: int = 0) -> str:
        ret = await self.async_command(
            f"ptz.cgi?action=getPresets&channel={channel}"
        )
        return ret.content.decode()

    @property
    def ptz_presets_count(self) -> int:
        return self.get_ptz_presets_count(channel=0)

    @property
    async def async_ptz_presets_count(self) -> int:
        return await self.async_get_ptz_presets_count(channel=0)

    def get_ptz_presets_count(self, *, channel: int = 0) -> int:
        ret = self.ptz_presets_list(channel=channel)
        return ret.count("Name=")

    async def async_get_ptz_presets_count(self, *, channel: int = 0) -> int:
        ret = await self.async_ptz_presets_list(channel=channel)
        return ret.count("Name=")

    def ptz_status(self, *, channel: int = 0) -> str:
        ret = self.command(f"ptz.cgi?action=getStatus&channel={channel}")
        return ret.content.decode()

    async def async_ptz_status(self, *, channel: int = 0) -> str:
        ret = await self.async_command(
            f"ptz.cgi?action=getStatus&channel={channel}"
        )
        return ret.content.decode()

    def ptz_tour_routines_list(self, *, channel: int = 0) -> str:
        ret = self.command(
            f"configManager.cgi?action=getTours&channel={channel}"
        )
        return ret.content.decode()

    async def async_ptz_tour_routines_list(self, *, channel: int = 0) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=getTours&channel={channel}"
        )
        return ret.content.decode()

    def ptz_control_command(
        self,
        action: ActionT,
        code: str,
        *,
        arg1: str = "0",
        arg2: str = "0",
        arg3: str = "0",
        channel: int = 0,
    ) -> bool:
        ret = self.command(
            f"ptz.cgi?action={action}&channel={channel}&code={code}&"
            f"arg1={arg1}&arg2={arg2}&arg3={arg3}"
        )
        return ret.content.decode().strip() == "OK"

    async def async_ptz_control_command(
        self,
        action: ActionT,
        code: str,
        *,
        arg1: str = "0",
        arg2: str = "0",
        arg3: str = "0",
        channel: int = 0,
    ) -> bool:
        ret = await self.async_command(
            f"ptz.cgi?action={action}&channel={channel}&code={code}&"
            f"arg1={arg1}&arg2={arg2}&arg3={arg3}"
        )
        return ret.content.decode().strip() == "OK"

    def zoom_in(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number

        The magic of zoom in 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="ZoomTele",
            channel=channel,
        )

    async def async_zoom_in(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number

        The magic of zoom in 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="ZoomTele",
            channel=channel,
        )

    def zoom_out(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number

        The magic of zoom out 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="ZoomWide",
            channel=channel,
        )

    async def async_zoom_out(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number

        The magic of zoom out 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="ZoomWide",
            channel=channel,
        )

    def move_left(
        self, start: bool, *, channel: int = 0, horizontal_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            horizontal_speed    - range 1-8

        The magic of move left 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="Left",
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_left(
        self, start: bool, *, channel: int = 0, horizontal_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            horizontal_speed    - range 1-8

        The magic of move left 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="Left",
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_right(
        self, start: bool, *, channel: int = 0, horizontal_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            horizontal_speed    - range 1-8

        The magic of move right 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="Right",
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_right(
        self, start: bool, *, channel: int = 0, horizontal_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            horizontal_speed    - range 1-8

        The magic of move right 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.5 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="Right",
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_up(
        self, start: bool, *, channel: int = 0, vertical_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            vertical_speed      - range 1-8

        The magic of move up 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.2 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="Up",
            arg2=str(vertical_speed),
            channel=channel,
        )

    async def async_move_up(
        self, start: bool, *, channel: int = 0, vertical_speed: int = 1
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
            vertical_speed      - range 1-8

        The magic of move up 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.2 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="Up",
            arg2=str(vertical_speed),
            channel=channel,
        )

    def move_down(
        self, start: bool, *, channel: int = 0, vertical_speed: int = 1
    ) -> bool:
        """
        The magic of move down 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.2 sec
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="Down",
            arg2=str(vertical_speed),
            channel=channel,
        )

    async def async_move_down(
        self, start: bool, *, channel: int = 0, vertical_speed: int = 1
    ) -> bool:
        """
        The magic of move down 1x, 2x etc. is the timer between the cmd
        'start' and cmd 'stop'. My suggestion for start/stop cmd is 0.2 sec
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="Down",
            arg2=str(vertical_speed),
            channel=channel,
        )

    def position_abs(
        self,
        start: bool,
        *,
        horizontal_angle: int = 0,
        vertical_angle: int = 0,
        channel: int = 0,
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            horizontal_angle    - range 0-360
            vertical_angle      - range 0-360
            channel             - channel number

        Go to an absolute coordinate
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="PositionABS",
            arg1=str(horizontal_angle),
            arg2=str(vertical_angle),
            channel=channel,
        )

    async def async_position_abs(
        self,
        start: bool,
        *,
        horizontal_angle: int = 0,
        vertical_angle: int = 0,
        channel: int = 0,
    ) -> bool:
        """
        Params:
            start               - True to start, False to stop
            horizontal_angle    - range 0-360
            vertical_angle      - range 0-360
            channel             - channel number

        Go to an absolute coordinate
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="PositionABS",
            arg1=str(horizontal_angle),
            arg2=str(vertical_angle),
            channel=channel,
        )

    def focus_near(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="FocusNear",
            channel=channel,
        )

    async def async_focus_near(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="FocusNear",
            channel=channel,
        )

    def focus_far(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="FocusFar",
            channel=channel,
        )

    async def async_focus_far(self, start: bool, *, channel: int = 0) -> bool:
        """
        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="FocusFar",
            channel=channel,
        )

    def iris_large(self, start: bool, *, channel: int = 0) -> bool:
        """
        Aperture larger

        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="IrisLarge",
            channel=channel,
        )

    async def async_iris_large(self, start: bool, *, channel: int = 0) -> bool:
        """
        Aperture larger

        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="IrisLarge",
            channel=channel,
        )

    def iris_small(self, start: bool, *, channel: int = 0) -> bool:
        """
        Aperture smaller

        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="IrisSmall",
            channel=channel,
        )

    async def async_iris_small(self, start: bool, *, channel: int = 0) -> bool:
        """
        Aperture smaller

        Params:
            start               - True to start, False to stop
            channel             - channel number
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="IrisSmall",
            channel=channel,
        )

    def go_to_preset(
        self, *, channel: int = 0, preset_point_number: int = 1
    ) -> bool:
        """
        Params:
            channel             - channel number
            preset_point_number - preset point number
        """
        return self.ptz_control_command(
            action="start",
            code="GotoPreset",
            arg2=str(preset_point_number),
            channel=channel,
        )

    async def async_go_to_preset(
        self, *, channel: int = 0, preset_point_number: int = 1
    ) -> bool:
        """
        Params:
            channel             - channel number
            preset_point_number - preset point number
        """
        return await self.async_ptz_control_command(
            action="start",
            code="GotoPreset",
            arg2=str(preset_point_number),
            channel=channel,
        )

    def set_preset(
        self, *, channel: int = 0, preset_point_number: int = 1
    ) -> bool:
        """
        Params:
            channel             - channel number
            preset_point_number - preset point number
        """
        return self.ptz_control_command(
            action="start",
            code="SetPreset",
            arg2=str(preset_point_number),
            channel=channel,
        )

    async def async_set_preset(
        self, *, channel: int = 0, preset_point_number: int = 1
    ) -> bool:
        """
        Params:
            channel             - channel number
            preset_point_number - preset point number
        """
        return await self.async_ptz_control_command(
            action="start",
            code="SetPreset",
            arg2=str(preset_point_number),
            channel=channel,
        )

    def tour(
        self,
        start: bool = True,
        *,
        channel: int = 0,
        tour_path_number: int = 1,
    ) -> bool:
        """
        Params:
            start               - True (StartTour) or False (StopTour)
            channel             - channel number
            tour_path_number    - tour path number
        """
        return self.ptz_control_command(
            action="start",
            code="StartTour" if start else "StopTour",
            arg1=str(tour_path_number),
            channel=channel,
        )

    async def async_tour(
        self,
        start: bool = True,
        *,
        channel: int = 0,
        tour_path_number: int = 1,
    ) -> bool:
        """
        Params:
            start               - True (StartTour) or False (StopTour)
            channel             - channel number
            tour_path_number    - tour path number
        """
        return await self.async_ptz_control_command(
            action="start",
            code="StartTour" if start else "StopTour",
            arg1=str(tour_path_number),
            channel=channel,
        )

    def move_left_up(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="LeftUp",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_left_up(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="LeftUp",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_left_down(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="LeftDown",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_left_down(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="LeftDown",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_right_up(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="RightUp",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_right_up(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="RightUp",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_right_down(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return self.ptz_control_command(
            action="start" if start else "stop",
            code="RightDown",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    async def async_move_right_down(
        self,
        start: bool,
        *,
        channel: int = 0,
        vertical_speed: int = 1,
        horizontal_speed: int = 1,
    ) -> bool:
        """
        Params:
            start            - True to start, False to stop
            channel          - channel number
            vertical_speed   - range is 1-8
            horizontal_speed - range is 1-8
        """
        return await self.async_ptz_control_command(
            action="start" if start else "stop",
            code="RightDown",
            arg1=str(vertical_speed),
            arg2=str(horizontal_speed),
            channel=channel,
        )

    def move_directly(
        self,
        startpoint_x: int,
        startpoint_y: int,
        endpoint_x: int,
        endpoint_y: int,
        *,
        channel: int = 1,
    ) -> str:
        """

        Three-dimensional orientation. Move to the rectangle with screen
        coordinate [startX, startY], [endX, endY]

        Params:
            channel          - channel index, start with 1
            startX, startY, endX and endY - range is 0-8192
        """
        ret = self.command(
            f"ptzBase.cgi?action=moveDirectly&channel={channel}&"
            f"startPoint[0]={startpoint_x}&startPoint[1]={startpoint_y}&"
            f"endPoint[0]={endpoint_x}&endPoint[1]={endpoint_y}"
        )
        return ret.content.decode()

    async def async_move_directly(
        self,
        startpoint_x: int,
        startpoint_y: int,
        endpoint_x: int,
        endpoint_y: int,
        *,
        channel: int = 1,
    ) -> str:
        """

        Three-dimensional orientation. Move to the rectangle with screen
        coordinate [startX, startY], [endX, endY]

        Params:
            channel          - channel index, start with 1
            startX, startY, endX and endY - range is 0-8192
        """
        ret = await self.async_command(
            f"ptzBase.cgi?action=moveDirectly&channel={channel}&"
            f"startPoint[0]={startpoint_x}&startPoint[1]={startpoint_y}&"
            f"endPoint[0]={endpoint_x}&endPoint[1]={endpoint_y}"
        )
        return ret.content.decode()



"""Test utils.py functions."""
from unittest import TestCase

from amcrest.utils import clean_url, percent, str2bool, to_unit


class TestUtils(TestCase):
    """Tests for utils.py."""

    def test_clean_url(self):
        """Test return code from utils.clean_url."""
        self.assertEqual('www.foo.bar', clean_url('http://www.foo.bar/'))
        self.assertEqual('www.foo.bar', clean_url('https://www.foo.bar'))

    def test_precision(self):
        """Test return code from utils.precision."""
        self.assertEqual("10.0", percent(10, 100))
        self.assertRaises(TypeError, percent, 'a', 100)

    def test_str2bool(self):
        """Test return code from utils.str2bool."""
        self.assertTrue(str2bool('y'))
        self.assertTrue(str2bool('Y'))
        self.assertTrue(str2bool('yes'))
        self.assertTrue(str2bool(1))
        self.assertFalse(str2bool('n'))
        self.assertFalse(str2bool('N'))
        self.assertFalse(str2bool('no'))
        self.assertFalse(str2bool(0))
        self.assertTrue(str2bool(u'Y'))
        self.assertRaises(ValueError, str2bool, 'amcrest')

    def test_to_unit(self):
        """Test return code from utils.to_unit."""
        self.assertEqual("1024.0", to_unit(1024, unit='B')[0])
        self.assertEqual("8192.0", to_unit(8589934592, unit='MB')[0])
        self.assertEqual("1024.0", to_unit(1099511627776, unit='GB')[0])
        self.assertNotEqual("117.73", to_unit('123000300', unit='MB')[0])



import responses
import amcrest
from unittest import TestCase

try:
    import urllib.parse
    fn_urlencode = urllib.parse.urlencode
except ImportError:
    import urllib
    fn_urlencode = urllib.urlencode


class MockTestCase(TestCase):
    def get_host(self):
        return 'www.example.com'

    def get_port(self):
        return 80

    def get_root_url(self):
        return 'http://{}:{}'.format(self.get_host(), self.get_port())

    def format_url(self, method, args=None):
        args_str = '?' + fn_urlencode(args) if args and len(args) > 0 else ''
        result = self.get_root_url() + '/cgi-bin/' + method + args_str
        return result

    def add_init_responses(self):
        responses.add(
            responses.GET,
            self.format_url('magicBox.cgi', {
                'action': 'getMachineName'}),
            body='name=AMCTEST_MACHINE',
            status=200)

        responses.add(
            responses.GET,
            self.format_url('magicBox.cgi', {
                'action': 'getSerialNo'}),
            body='sn=AMCTESTSERIALNO',
            status=200)

    def get_amcrest(self):
        return amcrest.AmcrestCamera(
            self.get_host(),
            self.get_port(),
            'admin',
            'test')



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import asyncio
import logging
import re
import socket
import threading
from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional, Tuple, Union

import httpx
import requests
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection

from .config import (
    KEEPALIVE_COUNT,
    KEEPALIVE_IDLE,
    KEEPALIVE_INTERVAL,
    MAX_RETRY_HTTP_CONNECTION,
    TIMEOUT_HTTP_PROTOCOL,
)
from .exceptions import CommError, LoginError, ReadTimeoutError
from .utils import clean_url, pretty

_LOGGER = logging.getLogger(__name__)

_KEEPALIVE_OPTS = HTTPConnection.default_socket_options + [
    (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
]
# On some systems TCP_KEEP* are not defined in socket.
try:
    _KEEPALIVE_OPTS += [
        (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, KEEPALIVE_IDLE),
        (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, KEEPALIVE_INTERVAL),
        (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, KEEPALIVE_COUNT),
    ]
except AttributeError:
    pass

TimeoutT = Union[Optional[float], Tuple[Optional[float], Optional[float]]]
HttpxTimeoutT = Union[
    Optional[float],
    Tuple[Optional[float], Optional[float], Optional[float], Optional[float]],
]


class SOHTTPAdapter(HTTPAdapter):
    """HTTPAdapter with support for socket options."""

    def __init__(self, *args, **kwargs) -> None:
        self.socket_options = kwargs.pop("socket_options", None)
        super().__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs) -> None:
        if self.socket_options is not None:
            kwargs["socket_options"] = self.socket_options
        super().init_poolmanager(*args, **kwargs)


class Http:
    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        *,
        verbose: bool = True,
        protocol: str = "http",
        ssl_verify: bool = True,
        retries_connection: Optional[int] = None,
        timeout_protocol: TimeoutT = None,
    ) -> None:
        self._token_lock = threading.Lock()
        try:
            self._async_token_lock: Optional[asyncio.Lock] = asyncio.Lock()
        except RuntimeError:
            self._async_token_lock = None
        self._cmd_id_lock = threading.Lock()
        self._cmd_id = 0
        self._host = clean_url(host)
        self._port = port
        self._user = user
        self._password = password
        self._verbose = verbose
        self._protocol = protocol
        self._verify = ssl_verify
        self._base_url = self.__base_url()

        self._retries_default = (
            retries_connection
            if retries_connection is not None
            else MAX_RETRY_HTTP_CONNECTION
        )
        self._timeout_default = timeout_protocol or TIMEOUT_HTTP_PROTOCOL

        self._token: Optional[requests.auth.AuthBase] = None
        self._async_token: Optional[httpx.Auth] = None
        self._name: Optional[str] = None
        self._serial: Optional[str] = None

    def _generate_token(self) -> None:
        """Create authentation to use with requests."""
        cmd = "magicBox.cgi?action=getMachineName"
        _LOGGER.debug("%s Trying Basic Authentication", self)
        self._token = requests.auth.HTTPBasicAuth(self._user, self._password)
        try:
            try:
                resp = self._command(cmd).content.decode()
            except LoginError:
                _LOGGER.debug("%s Trying Digest Authentication", self)
                self._token = requests.auth.HTTPDigestAuth(
                    self._user, self._password
                )
                resp = self._command(cmd).content.decode()
        except CommError:
            self._token = None
            raise

        # check if user passed
        result = resp.lower()
        if "invalid" in result or "error" in result:
            _LOGGER.debug(
                "%s Result from camera: %s",
                self,
                resp.strip().replace("\r\n", ": "),
            )
            self._token = None
            raise LoginError("Invalid credentials")

        self._name = pretty(resp.strip())

        _LOGGER.debug("%s Retrieving serial number", self)
        self._serial = pretty(
            self._command("magicBox.cgi?action=getSerialNo")
            .content.decode()
            .strip()
        )

    async def _async_generate_token(self) -> None:
        """Create authentation to use with requests."""
        cmd = "magicBox.cgi?action=getMachineName"
        _LOGGER.debug("%s Trying async Basic Authentication", self)
        self._async_token = httpx.BasicAuth(self._user, self._password)
        try:
            try:
                resp = (await self._async_command(cmd)).content.decode()
            except LoginError:
                _LOGGER.debug("%s Trying async Digest Authentication", self)
                self._async_token = httpx.DigestAuth(
                    self._user, self._password
                )
                resp = (await self._async_command(cmd)).content.decode()
        except CommError:
            self._async_token = None
            raise

        # check if user passed
        result = resp.lower()
        if "invalid" in result or "error" in result:
            _LOGGER.debug(
                "%s Result from camera: %s",
                self,
                resp.strip().replace("\r\n", ": "),
            )
            self._async_token = None
            raise LoginError("Invalid credentials")

        self._name = pretty(resp.strip())

        _LOGGER.debug("%s Retrieving serial number", self)
        self._serial = pretty(
            (await self._async_command("magicBox.cgi?action=getSerialNo"))
            .content.decode()
            .strip()
        )

    def __repr__(self) -> str:
        """Default object representation."""
        if self._name is None:
            return f"<Unconnected @ {self._host}>"
        if self._serial is None:
            return f"<{self._name}:CONNECTING>"
        return f"<{self._name}:{self._serial}>"

    def as_dict(self) -> dict:
        """Callback for __dict__."""
        cdict = self.__dict__.copy()
        redacted = "**********"
        if cdict["_token"] is not None:
            cdict["_token"] = redacted
        cdict["_password"] = redacted
        return cdict

    # Base methods
    def __base_url(self, param: str = "") -> str:
        return f"{self._protocol}://{self._host}:{self._port}/cgi-bin/{param}"

    def get_base_url(self) -> str:
        return self._base_url

    def command(self, *args, **kwargs) -> requests.Response:
        """
        Args:
            cmd - command to execute via http
            retries - maximum number of retries each connection should attempt
            timeout_cmd - timeout
            stream - if True do not download entire response immediately
        """
        with self._token_lock:
            if not self._token:
                self._generate_token()
        return self._command(*args, **kwargs)

    async def async_command(self, *args, **kwargs) -> httpx.Response:
        if self._async_token_lock is None:
            raise RuntimeError("Camera not setup with async loop running")
        async with self._async_token_lock:
            if not self._async_token:
                await self._async_generate_token()
        return await self._async_command(*args, **kwargs)

    @asynccontextmanager
    async def async_stream_command(
        self, *args, **kwargs
    ) -> AsyncIterator[httpx.Response]:
        if self._async_token_lock is None:
            raise RuntimeError("Camera not setup with async loop running")
        async with self._async_token_lock:
            if not self._async_token:
                await self._async_generate_token()
        async with self._async_stream_command(*args, **kwargs) as ret:
            yield ret

    def _command(
        self,
        cmd: str,
        retries: Optional[int] = None,
        timeout_cmd: TimeoutT = None,
        stream: bool = False,
    ) -> requests.Response:
        url = self.__base_url(cmd)
        with self._cmd_id_lock:
            cmd_id = self._cmd_id = self._cmd_id + 1
        _LOGGER.debug("%s HTTP query %i: %s", self, cmd_id, url)
        if retries is None:
            retries = self._retries_default
        timeout = timeout_cmd or self._timeout_default
        with requests.Session() as session:
            if isinstance(timeout, float):
                use_keepalive = False
            else:
                use_keepalive = timeout[0] is None or timeout[1] is None
            if use_keepalive:
                session.mount(
                    "{0}://".format(self._protocol),
                    SOHTTPAdapter(socket_options=_KEEPALIVE_OPTS),
                )
            for loop in range(1, 2 + retries):
                _LOGGER.debug(
                    "%s Running query %i attempt %s", self, cmd_id, loop
                )
                try:
                    resp = session.get(
                        url,
                        auth=self._token,
                        stream=stream,
                        timeout=timeout,
                        verify=self._verify,
                    )
                    if resp.status_code == 401:
                        _LOGGER.debug(
                            "%s Query %i: Unauthorized (401)", self, cmd_id
                        )
                        self._token = None
                        raise LoginError()
                    resp.raise_for_status()
                except requests.RequestException as error:
                    _LOGGER.debug(
                        "%s Query %i failed due to error: %r",
                        self,
                        cmd_id,
                        error,
                    )
                    if loop > retries:
                        raise CommError(error) from error
                    msg = re.sub(
                        r"at 0x[0-9a-fA-F]+", "at ADDRESS", repr(error)
                    )
                    _LOGGER.warning(
                        "%s Trying again due to error: %s", self, msg
                    )
                    continue
                else:
                    break

        _LOGGER.debug(
            "%s Query %i worked. Exit code: <%s>",
            self,
            cmd_id,
            resp.status_code,
        )
        return resp

    async def _async_command(
        self,
        cmd: str,
        retries: Optional[int] = None,
        timeout_cmd: TimeoutT = None,
    ) -> httpx.Response:
        url = self.__base_url(cmd)
        cmd_id = self._cmd_id = self._cmd_id + 1
        _LOGGER.debug("%s Async HTTP query %i: %s", self, cmd_id, url)
        if retries is None:
            retries = self._retries_default

        timeout = timeout_cmd or self._timeout_default
        # requests uses (connect timeout, read timeout), and httpx adds (write
        # timeout, pool timeout), just set the extras to None
        if isinstance(timeout, tuple):
            httpx_timeout: HttpxTimeoutT = timeout + (None, None)
        else:
            httpx_timeout = timeout

        async with httpx.AsyncClient(
            follow_redirects=True,
            auth=self._async_token,
            verify=self._verify,
            timeout=httpx_timeout,
        ) as client:
            for loop in range(1, 2 + retries):
                _LOGGER.debug(
                    "%s Running query %i attempt %s", self, cmd_id, loop
                )
                try:
                    resp = await client.get(url)
                    if resp.status_code == 401:
                        _LOGGER.debug(
                            "%s Query %i: Unauthorized (401)", self, cmd_id
                        )
                        self._async_token = None
                        raise LoginError()
                    resp.raise_for_status()
                except httpx.HTTPError as error:
                    _LOGGER.debug(
                        "%s Query %i failed due to error: %r",
                        self,
                        cmd_id,
                        error,
                    )
                    if loop > retries:
                        raise CommError(error) from error
                    msg = re.sub(
                        r"at 0x[0-9a-fA-F]+", "at ADDRESS", repr(error)
                    )
                    _LOGGER.warning(
                        "%s Trying again due to error: %s", self, msg
                    )
                    continue
                else:
                    break

        _LOGGER.debug(
            "%s Query %i worked. Exit code: <%s>",
            self,
            cmd_id,
            resp.status_code,
        )
        return resp

    @asynccontextmanager
    async def _async_stream_command(
        self,
        cmd: str,
        timeout_cmd: TimeoutT = None,
    ) -> AsyncIterator[httpx.Response]:
        url = self.__base_url(cmd)
        cmd_id = self._cmd_id = self._cmd_id + 1
        _LOGGER.debug("%s HTTP query %i: %s", self, cmd_id, url)

        timeout = timeout_cmd or self._timeout_default
        # requests uses (connect timeout, read timeout), and httpx adds (write
        # timeout, pool timeout), just set the extras to None
        if isinstance(timeout, tuple):
            httpx_timeout: HttpxTimeoutT = timeout + (None, None)
        else:
            httpx_timeout = timeout

        async with httpx.AsyncClient(
            follow_redirects=True,
            auth=self._async_token,
            verify=self._verify,
            timeout=httpx_timeout,
        ) as client:
            try:
                async with client.stream("GET", url) as resp:
                    if resp.status_code == 401:
                        _LOGGER.debug(
                            "%s Query %i: Unauthorized (401)", self, cmd_id
                        )
                        self._async_token = None
                        raise LoginError()
                    resp.raise_for_status()

                    _LOGGER.debug(
                        "%s Query %i worked. Exit code: <%s>",
                        self,
                        cmd_id,
                        resp.status_code,
                    )
                    yield resp
            except httpx.HTTPError as error:
                _LOGGER.debug(
                    "%s Query %i failed due to error: %r",
                    self,
                    cmd_id,
                    error,
                )
                if isinstance(error, httpx.ReadTimeout):
                    raise ReadTimeoutError(error) from error
                else:
                    raise CommError(error) from error

    def command_audio(
        self,
        cmd: str,
        file_content,
        http_header,
        timeout: TimeoutT = None,
    ) -> None:
        with self._token_lock:
            if not self._token:
                self._generate_token()
        url = self.__base_url(cmd)
        try:
            requests.post(
                url,
                files=file_content,
                auth=self._token,
                headers=http_header,
                timeout=timeout or self._timeout_default,
            )
        except requests.exceptions.ReadTimeout:
            pass

    # Helpers for common commands

    def _get_config(self, config_name: str) -> str:
        ret = self.command(
            f"configManager.cgi?action=getConfig&name={config_name}"
        )
        return ret.content.decode()

    async def _async_get_config(self, config_name: str) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=getConfig&name={config_name}"
        )
        return ret.content.decode()

    def _magic_box(self, action: str) -> str:
        ret = self.command(f"magicBox.cgi?action={action}")
        return ret.content.decode()

    async def _async_magic_box(self, action: str) -> str:
        ret = await self.async_command(f"magicBox.cgi?action={action}")
        return ret.content.decode()



"""Test log.py functions."""
import responses
import datetime

from .mocktestcase import MockTestCase


class TestMedia(MockTestCase):
    """Tests for media.py."""
    @responses.activate
    def test_download_file(self):
        self.add_init_responses()

        body = '*'.join([str(x) for x in range(17)])

        responses.add(
            responses.GET,
            self.format_url('RPC_Loadfile//mnt/sd/test.dat'),
            body=body,
            status=200)

        c = self.get_amcrest().camera
        content = c.download_file('/mnt/sd/test.dat').decode('utf-8')
        self.assertEqual(body, content)

    @responses.activate
    def test_log_find(self):
        self.add_init_responses()

        responses.add(
            responses.GET,
            self.format_url('mediaFileFind.cgi', {
                'action': 'factory.create'}),
            body='result=123\r\n',
            status=200)

        responses.add(
            responses.GET,
            self.format_url('mediaFileFind.cgi', {
                'action': 'findFile',
                'object': 123,
                'condition.Channel': 0,
                'condition.StartTime': datetime.datetime(2020, 1, 2, 3, 4, 5),
                'condition.EndTime': datetime.datetime(2020, 1, 2, 3, 4, 10)}),
            body='OK\r\n',
            status=200)

        url_next = self.format_url('mediaFileFind.cgi', {
            'action': 'findNextFile',
            'object': 123,
            'count': 100})

        responses.add(
            responses.GET,
            url_next,
            body='\r\n'.join('''found=2
items[0].Channel=0
items[0].Cluster=0
items[0].Compressed=false
items[0].Disk=0
items[0].Duration=480
items[0].EndTime=2020-01-06 06:13:00
items[0].Events[0]=VideoMotion
items[0].FilePath=/mnt/sd/2020-01-06/001/dav/06/06.05.00-06.13.00[M][0@0][0].mp4
items[0].Flags[0]=Event
items[0].Length=248362892
items[0].Overwrites=0
items[0].Partition=0
items[0].Redundant=false
items[0].Repeat=0
items[0].StartTime=2020-01-06 06:05:00
items[0].Summary.TrafficCar.PlateColor=Yellow
items[0].Summary.TrafficCar.PlateNumber=
items[0].Summary.TrafficCar.PlateType=Yellow
items[0].Summary.TrafficCar.Speed=60
items[0].Summary.TrafficCar.VehicleColor=White
items[0].SummaryOffset=0
items[0].Type=dav
items[0].WorkDir=/mnt/sd
items[0].WorkDirSN=0
items[1].Channel=0
items[1].Cluster=0
items[1].Compressed=false
items[1].Disk=0
items[1].Duration=480
items[1].EndTime=2020-01-06 06:21:00
items[1].Events[0]=VideoMotion
items[1].FilePath=/mnt/sd/2020-01-06/001/dav/06/06.13.00-06.21.00[M][0@0][0].mp4
items[1].Flags[0]=Event
items[1].Length=241799877
items[1].Overwrites=0
items[1].Partition=0
items[1].Redundant=false
items[1].Repeat=0
items[1].StartTime=2020-01-06 06:13:00
items[1].Summary.TrafficCar.PlateColor=Yellow
items[1].Summary.TrafficCar.PlateNumber=
items[1].Summary.TrafficCar.PlateType=Yellow
items[1].Summary.TrafficCar.Speed=60
items[1].Summary.TrafficCar.VehicleColor=White
items[1].SummaryOffset=0
items[1].Type=dav
items[1].WorkDir=/mnt/sd
items[1].WorkDirSN=0

'''.splitlines()),
            status=200)

        responses.add(
            responses.GET,
            url_next,
            body='\r\n'.join('''found=2
items[0].Channel=0
items[0].Cluster=0
items[0].Compressed=false
items[0].Disk=0
items[0].Duration=34
items[0].EndTime=2020-01-06 06:36:44
items[0].Events[0]=VideoMotion
items[0].FilePath=/mnt/sd/2020-01-06/001/dav/06/06.36.10-06.36.44[M][0@0][0].mp4
items[0].Flags[0]=Event
items[0].Length=17561629
items[0].Overwrites=0
items[0].Partition=0
items[0].Redundant=false
items[0].Repeat=0
items[0].StartTime=2020-01-06 06:36:10
items[0].Summary.TrafficCar.PlateColor=Yellow
items[0].Summary.TrafficCar.PlateNumber=
items[0].Summary.TrafficCar.PlateType=Yellow
items[0].Summary.TrafficCar.Speed=60
items[0].Summary.TrafficCar.VehicleColor=White
items[0].SummaryOffset=0
items[0].Type=dav
items[0].WorkDir=/mnt/sd
items[0].WorkDirSN=0

'''.splitlines()),
            status=200)

        responses.add(
            responses.GET,
            url_next,
            body='\r\n'.join('''found=0

'''.splitlines()),
            status=200)

        responses.add(
            responses.GET,
            self.format_url('mediaFileFind.cgi', {
                'action': 'factory.close',
                'object': 123}),
            body='\r\n'.join('''Error
ErrorID=2, Detail=Invalid Request!

'''.splitlines()),
            status=200)

        responses.add(
            responses.GET,
            self.format_url('mediaFileFind.cgi', {
                'action': 'factory.destroy',
                'object': 123}),
            body='\r\n'.join('''Error
ErrorID=2, Detail=Invalid Request!

'''.splitlines()),
            status=200)

        c = self.get_amcrest().camera
        time_start = datetime.datetime(2020, 1, 2, 3, 4, 5)
        time_end = datetime.datetime(2020, 1, 2, 3, 4, 10)
        media = list(c.find_files(time_start, time_end))
        self.assertEqual(2, len(media))



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from amcrest.http import Http
from amcrest.utils import pretty, str2bool


class MotionDetection(Http):
    @property
    def motion_detection(self) -> str:
        return self._get_config("MotionDetect")

    @motion_detection.setter
    def motion_detection(self, opt: bool) -> bool:
        return self.set_motion_detection(opt)

    @motion_detection.setter  # type: ignore[attr-defined]
    def motion_recording(self, opt: bool) -> bool:
        return self.set_motion_recording(opt)

    @property
    async def async_motion_detection(self) -> str:
        return await self._async_get_config("MotionDetect")

    def is_motion_detector_on(self, *, channel: int = 0) -> bool:
        ret = self.motion_detection
        status = [pretty(s) for s in ret.split() if ".Enable=" in s][channel]
        return str2bool(status)

    async def async_is_motion_detector_on(self, *, channel: int = 0) -> bool:
        ret = await self.async_motion_detection
        status = [pretty(s) for s in ret.split() if ".Enable=" in s][channel]
        return str2bool(status)

    def is_record_on_motion_detection(self, *, channel: int = 0) -> bool:
        ret = self.motion_detection
        status = [pretty(s) for s in ret.split() if ".RecordEnable=" in s][
            channel
        ]
        return str2bool(status)

    async def async_is_record_on_motion_detection(
        self, *, channel: int = 0
    ) -> bool:
        ret = await self.async_motion_detection
        status = [pretty(s) for s in ret.split() if ".RecordEnable=" in s][
            channel
        ]
        return str2bool(status)

    def set_motion_detection(self, opt: bool, *, channel: int = 0) -> bool:
        value = "true" if opt else "false"
        ret = self.command(
            "configManager.cgi?action="
            f"setConfig&MotionDetect[{channel}].Enable={value}"
        )
        return "ok" in ret.content.decode().lower()

    async def async_set_motion_detection(
        self, opt: bool, *, channel: int = 0
    ) -> bool:
        value = "true" if opt else "false"
        ret = await self.async_command(
            "configManager.cgi?action="
            f"setConfig&MotionDetect[{channel}].Enable={value}"
        )
        return "ok" in ret.content.decode().lower()

    def set_motion_recording(self, opt: bool, *, channel: int = 0) -> bool:
        value = "true" if opt else "false"
        ret = self.command(
            "configManager.cgi?action="
            f"setConfig&MotionDetect[{channel}].EventHandler."
            f"RecordEnable={value}"
        )
        return "ok" in ret.content.decode().lower()

    async def async_set_motion_recording(
        self, opt: bool, *, channel: int = 0
    ) -> bool:
        value = "true" if opt else "false"
        ret = await self.async_command(
            "configManager.cgi?action="
            f"setConfig&MotionDetect[{channel}].EventHandler."
            f"RecordEnable={value}"
        )
        return "ok" in ret.content.decode().lower()



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

from amcrest import AmcrestCamera

amcrest = AmcrestCamera('192.168.1.10', 80, 'admin', 'super_password')
camera = amcrest.camera

camera.snapshot(channel=0, path_file="/tmp/snapshot.jpg")



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et

import datetime
from amcrest import AmcrestCamera

amcrest = AmcrestCamera("192.168.1.10", 80, "admin", "super_password")
camera = amcrest.camera

end_time = datetime.datetime.now()
start_time = end_time - datetime.timedelta(hours=1)

for text in camera.log_find(start_time, end_time):
    print(text)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import socket
import threading
from contextlib import closing
from typing import List, Optional

from .http import Http


class Network(Http):

    amcrest_ips: List[str] = []
    __RTSP_PORT = 554
    __PWGPSI_PORT = 3800

    def __raw_scan(self, ipaddr: str, timeout: Optional[float] = None) -> None:
        if timeout:
            socket.setdefaulttimeout(timeout)
        else:
            # If devices not found, try increasing timeout
            socket.setdefaulttimeout(0.2)

        with closing(socket.socket()) as sock:
            try:
                sock.connect((ipaddr, self.__RTSP_PORT))
                sock.connect((ipaddr, self.__PWGPSI_PORT))
                self.amcrest_ips.append(ipaddr)

            # pylint: disable=bare-except
            except:
                pass

    def scan_devices(
        self, subnet: str, timeout: Optional[float] = None
    ) -> List[str]:
        """
        Scan cameras in a range of ips

        Params:
        subnet - subnet, i.e: 192.168.1.0/24
                 if mask not used, assuming mask 24

        timeout - timeout in sec

        Returns:
        """

        # Maximum range from mask
        # Format is mask: max_range
        max_range = {
            16: 256,
            24: 256,
            25: 128,
            27: 32,
            28: 16,
            29: 8,
            30: 4,
            31: 2,
        }

        # If user didn't provide mask, use /24
        if "/" not in subnet:
            mask = int(24)
            network = subnet
        else:
            network, mask_str = subnet.split("/")
            mask = int(mask_str)

        if mask not in max_range:
            raise RuntimeError("Cannot determine the subnet mask!")

        # Default logic is remove everything from last "." to the end
        # This logic change in case mask is 16
        network = network.rpartition(".")[0]

        if mask == 16:
            # For mask 16, we must cut the last two
            # entries with .

            # pylint: disable=unused-variable
            for i in range(0, 1):
                network = network.rpartition(".")[0]

        # Trigger the scan
        # For clear coding, let's keep the logic in if/else (mask16)
        # instead of only one if
        if mask == 16:
            for seq1 in range(max_range[mask]):
                for seq2 in range(max_range[mask]):
                    ipaddr = f"{network}.{seq1}.{seq2}"
                    thd = threading.Thread(
                        target=self.__raw_scan, args=(ipaddr, timeout)
                    )
                    thd.start()
        else:
            for seq1 in range(max_range[mask]):
                ipaddr = f"{network}.{seq1}"
                thd = threading.Thread(
                    target=self.__raw_scan, args=(ipaddr, timeout)
                )
                thd.start()

        return self.amcrest_ips

    @property
    def wlan_config(self) -> str:
        return self._get_config("WLan")

    @property
    async def async_wlan_config(self) -> str:
        return await self._async_get_config("WLan")

    @property
    def telnet_config(self) -> str:
        return self._get_config("Telnet")

    @telnet_config.setter
    def telnet_config(self, status: str) -> str:
        """
        status:
            false - Telnet is disabled
            true  - Telnet is enabled
        """
        ret = self.command(
            f"configManager.cgi?action=setConfig&Telnet.Enable={status}"
        )
        return ret.content.decode()

    @property
    async def async_telnet_config(self) -> str:
        return await self._async_get_config("Telnet")

    async def async_set_telnet_config(self, status: str) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=setConfig&Telnet.Enable={status}"
        )
        return ret.content.decode()

    @property
    def network_config(self) -> str:
        return self._get_config("Network")

    @property
    async def async_network_config(self) -> str:
        return await self._async_get_config("Network")

    @property
    def network_interfaces(self) -> str:
        ret = self.command("netApp.cgi?action=getInterfaces")
        return ret.content.decode()

    @property
    async def async_network_interfaces(self) -> str:
        ret = await self.async_command("netApp.cgi?action=getInterfaces")
        return ret.content.decode()

    @property
    def upnp_status(self) -> str:
        ret = self.command("netApp.cgi?action=getUPnPStatus")
        return ret.content.decode()

    @property
    async def async_upnp_status(self) -> str:
        ret = await self.async_command("netApp.cgi?action=getUPnPStatus")
        return ret.content.decode()

    @property
    def upnp_config(self) -> str:
        return self._get_config("UPnP")

    @upnp_config.setter
    def upnp_config(self, upnp_opt: str) -> str:
        """
        01/21/2017

        Note 1:
        -------
        The current SDK from Amcrest is case sensitive, do not
        mix UPPERCASE options with lowercase. Otherwise it will
        ignore your call.

        Example:

        Correct:
                "UPnP.Enable=true&UPnP.MapTable[0].Protocol=UDP"

        Incorrect:
            "UPnP.Enable=true&UPnP.Maptable[0].Protocol=UDP"
                                      ^ here should be T in UPPERCASE

        Note 2:
        -------
        In firmware Amcrest_IPC-AWXX_Eng_N_V2.420.AC00.15.R.20160908.bin
        InnerPort was not able to be changed as API SDK 2.10 suggests.

        upnp_opt is the UPnP options listed as example below:
        +-------------------------------------------------------------------+
        | ParamName                      | Value  | Description             |
        +--------------------------------+----------------------------------+
        |UPnP.Enable                     | bool   | Enable/Disable UPnP     |
        |UPnP.MapTable[index].Enable     | bool   | Enable/Disable UPnP map |
        |UPnP.MapTable[index].InnerPort  | int    | Range [1-65535]         |
        |UPnP.MapTable[index].OuterPort  | int    | Range [1-65535]         |
        |UPnP.MapTable[index].Protocol   | string | Range {TCP, UDP}        |
        |UPnP.MapTable[index].ServiceName| string | User UPnP Service name  |
        +-------------------------------------------------------------------+

        upnp_opt format:
        <paramName>=<paramValue>[&<paramName>=<paramValue>...]
        """
        ret = self.command(f"configManager.cgi?action=setConfig&{upnp_opt}")
        return ret.content.decode()

    @property
    async def async_upnp_config(self) -> str:
        return await self._async_get_config("UPnP")

    async def async_set_upnp_config(self, upnp_opt: str) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=setConfig&{upnp_opt}"
        )
        return ret.content.decode()

    @property
    def ntp_config(self) -> str:
        return self._get_config("NTP")

    @ntp_config.setter
    def ntp_config(self, ntp_opt: str) -> str:
        """
        ntp_opt is the NTP options listed as example below:

        NTP.Address=clock.isc.org
        NTP.Enable=false
        NTP.Port=38
        NTP.TimeZone=9
        NTP.UpdatePeriod=31

        ntp_opt format:
        <paramName>=<paramValue>[&<paramName>=<paramValue>...]
        """
        ret = self.command(f"configManager.cgi?action=setConfig&{ntp_opt}")
        return ret.content.decode()

    @property
    async def async_ntp_config(self) -> str:
        return await self._async_get_config("NTP")

    async def async_set_ntp_config(self, ntp_opt: str) -> str:
        ret = await self.async_command(
            f"configManager.cgi?action=setConfig&{ntp_opt}"
        )
        return ret.content.decode()

    @property
    def rtsp_config(self) -> str:
        """Get RTSP configuration."""
        return self._get_config("RTSP")

    @property
    async def async_rtsp_config(self) -> str:
        """Get RTSP configuration."""
        return await self._async_get_config("RTSP")
